# Generated by Django 3.0.3 on 2020-08-25 11:35

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0005_auto_20200825_1128'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='labels',
            name='complete',
        ),
    ]
