        <link rel="stylesheet" href="/static/bootstrap-4.3.1.min.css">
        <link rel="stylesheet" href="/static/open-iconic/css/open-iconic-bootstrap.css">

        <link rel="stylesheet" href="/static/image_labeller.css">
