        <link rel="stylesheet" href="/static/bootstrap-4.3.1.min.css">
        <link rel="stylesheet" href="/static/jquery-ui-1.12.1.custom/jquery-ui.min.css">
        <link rel="stylesheet" href="/static/open-iconic/css/open-iconic-bootstrap.css">

        <link rel="stylesheet" href="/static/spectrum.css">
        <link rel="stylesheet" href="/static/schema_editor/schema_editor.css">
