        <script src="/static/jquery-3.6.0.min.js"></script>
        <script src="/static/js.cookie.js"></script>
        <script src="/static/json2.js"></script>

        <script src="/static/bootstrap-4.3.1.min.js"></script>
        <script src="/static/popper-1.14.7.min.js"></script>

        <script src="/static/spectrum.js"></script>
        <script src="/static/Sortable-1.10.0-rc3.js"></script>

        <script src="/static/vue/vue.global.js"></script>
        <script src="/static/vue/vuedraggable.umd.min.js"></script>

        <script src="/static/labelling_tool/object_id_table.js"></script>
        <script src="/static/schema_editor/schema_editor.js"></script>
