html, body {
    min-width: 100% !important;
    width: 100%;
    min-height: 100% !important;
    height: 100%;
    overflow: hidden;
}

div.image_labeller {
    width: inherit;
    height: inherit;
    overflow: hidden;
}

nav.labeller_menu_bar {
    padding: 3px;
    width: inherit;
    border-bottom: #5c637a 1px solid;
    background-color: #383c4a;
    color: #e8ecf0;
}

.labeller_menu_bar_right {
    position: fixed;
    display: inline-block;
    right: 0;
}

.labeller_menu_bar_right > h4 {
    display: inline-block;
}

.labeller_menu_sep {
    display: inline-block;
    width: 50px;
}


.tool_bar_popup {
    background-color: #ffffff;
    border: 1px solid rgba(0,0,0,.15);
    border-radius: .25rem;
    padding: 16px 24px;
    margin: .125rem 0 0;
    max-width: 1280px;
    max-height: 100%;
    overflow-y: auto;
}

.class_selector_group {
    font-size: .875rem;
    color: #6c757d;
}


aside.tool_bar {
    position: fixed;
    background-color: #51576b;
    width: 300px;
    height: inherit;
    margin: 0;
    border-right: #727b97 1px solid;
    overflow: hidden;
}

.tool_bar_header {
    /*background-color: #5b6177;*/
    /*margin: 4px 4px;*/
    /*border-left: #727b97 15px solid;*/
    /*border-right: #727b97 15px solid;*/
    /*padding: 0 4px;*/
    /*text-align: center;*/
    background-color: #3f4457;
    margin: 4px 0;
    border-top: #62687d 1px solid;
    border-bottom: #62687d 1px solid;
    padding: 0 0;
    text-align: center;
}

.tool_bar_tool_header {
    /*background-color: #5b6177;*/
    /*margin: 4px 4px;*/
    /*border-left: #727b97 15px solid;*/
    /*border-right: #727b97 15px solid;*/
    /*padding: 0 4px;*/
    /*text-align: center;*/
    background-color: #415252;
    margin: 4px 0;
    border-top: #627575 1px solid;
    border-bottom: #627575 1px solid;
    padding: 0 0;
    text-align: center;
}

.tool_bar_group {
    width: inherit;
}

.tool_bar_content {
    width: inherit;
    padding: 0 10px 0 10px;
    margin: 1px 0;
}


.tool_canvas_container {
    position: relative;
    width: inherit;
    height: inherit;
    margin-left: 300px;
    overflow: hidden;
    background-color: #22242d;
}

.tool_canvas {
    width: inherit;
    height: inherit;
}

.tool_canvas .anno_label {
    vector-effect: non-scaling-stroke;
}

.anno_hidden {
    display: none;
}

.anno_lockable {
}


.image-chooser-control {
    display: inline-block;
    width: 15em;
    vertical-align: middle;
}
