# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2017-04-11 07:40
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='labels',
            name='edit_time_elapsed',
            field=models.FloatField(blank=True, default=0.0),
        ),
    ]
