# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2020-02-11 18:20
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('example_labeller', '0002_auto_20180911_0954'),
    ]

    operations = [
        migrations.CreateModel(
            name='DextrTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_timestamp', models.DateTimeField(default=django.utils.timezone.now)),
                ('image_id_str', models.CharField(max_length=128)),
                ('dextr_id', models.IntegerField()),
                ('celery_task_id', models.CharField(max_length=128)),
                ('image', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='example_labeller.ImageWithLabels')),
            ],
        ),
    ]
