# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-30 09:03
from __future__ import unicode_literals

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Labels',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('labels_json_str', models.TextField(default=b'[]')),
                ('complete', models.BooleanField(default=False)),
                ('creation_date', models.DateField()),
                ('last_modified_datetime', models.DateTimeField(default=datetime.datetime.now)),
                ('lock_expiry_datetime', models.DateTimeField(default=datetime.datetime.now)),
                ('last_modified_by', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='modified_labels', to=settings.AUTH_USER_MODEL)),
                ('locked_by', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='locked_labels', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]
