# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-11 08:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0002_labels_edit_time_elapsed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='labels',
            name='labels_json_str',
            field=models.TextField(default='[]'),
        ),
    ]
