# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-30 09:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('image_labelling_tool', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ImageWithLabels',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(blank=True, upload_to=b'')),
                ('labels', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='image', to='image_labelling_tool.Labels')),
            ],
        ),
    ]
