        <script src="/static/jquery-3.6.0.min.js"></script>
        <script src="/static/popper-1.14.7.min.js"></script>
        <script src="/static/bootstrap-4.3.1.min.js"></script>

        <script src="/static/js.cookie.js"></script>
        <script src="/static/json2.js"></script>

        <script src="/static/d3.min.js"></script>
        <script src="/static/polyk.js"></script>
        <script src="/static/polybool.min.js"></script>
        <script src="/static/socket.io.min.js"></script>

        <script src="/static/labelling_tool/math_primitives.js"></script>
        <script src="/static/labelling_tool/object_id_table.js"></script>
        <script src="/static/labelling_tool/schema.js"></script>
        <script src="/static/labelling_tool/abstract_label.js"></script>
        <script src="/static/labelling_tool/abstract_tool.js"></script>
        <script src="/static/labelling_tool/select_tools.js"></script>
        <script src="/static/labelling_tool/point_label.js"></script>
        <script src="/static/labelling_tool/box_label.js"></script>
        <script src="/static/labelling_tool/oriented_ellipse_label.js"></script>
        <script src="/static/labelling_tool/composite_label.js"></script>
        <script src="/static/labelling_tool/polygonal_label.js"></script>
        <script src="/static/labelling_tool/dextr_label.js"></script>
        <script src="/static/labelling_tool/group_label.js"></script>
        <script src="/static/labelling_tool/main_labeller.js"></script>
        <script src="/static/labelling_tool/root_label_view.js"></script>
        <script src="/static/labelling_tool/popup_menu.js"></script>
        <script src="/static/labelling_tool/anno_controls.js"></script>

        <script type="application/javascript">
            $(function () {
                $('[data-toggle="tooltip"]').tooltip({delay: 750});
            });
        </script>
